# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

import numpy as np

import xobjects as xo
import xtrack as xt

from ..fieldmaps import TriLinearInterpolatedFieldMap
from ..fieldmaps import TriCubicInterpolatedFieldMap

from ..fieldmaps import TriCubicInterpolatedFieldMapData
from ..fieldmaps import TriLinearInterpolatedFieldMapData
from ..general import _pkg_root

class ElectronLensInterpolated(xt.BeamElement):

    _xofields={
               'current':  xo.Float64,
               'length':   xo.Float64,
               'voltage':  xo.Float64,
               "fieldmap": TriCubicInterpolatedFieldMapData,
              }

    def __init__(self,
                 _context=None,
                 _buffer=None,
                 _offset=None,
                 length=None,
                 fieldmap=None,
                 x_range=None, y_range=None,
                 nx=None, ny=None,
                 dx=None, dy=None,
                 x_grid=None, y_grid=None,
                 rho=None,
                 current=None, voltage=None,
                 ):

        if _buffer is not None:
            _context = _buffer.context
        if _context is None:
            _context = xo.context_default

        # Choose a small number of slices with a large range in zeta.
        # By copying the same potential on each slice, the 
        # interpolation becomes 2D, doesn't depend on the 
        # longitudinal variable.
        nz = 11 
        z_range=(-1,1) 

        fieldmap = TriLinearInterpolatedFieldMap(x_range=x_range, y_range=y_range, 
                                z_range=z_range, nx=nx, ny=ny, nz=nz, 
                                dx=dx, dy=dy, 
                                x_grid=x_grid, y_grid=y_grid,
                                solver="FFTSolver2p5D")

        for ii in range(nz):
            fieldmap.rho[:,:,ii] = rho
        fieldmap.update_phi_from_rho()

        tc_fieldmap = TriCubicInterpolatedFieldMap(x_grid=fieldmap._x_grid, 
                                                   y_grid=fieldmap._y_grid, 
                                                   z_grid=fieldmap._z_grid,
                                                  )

        nx = tc_fieldmap.nx
        ny = tc_fieldmap.ny
        nz = tc_fieldmap.nz
        dx = tc_fieldmap.dx 
        dy = tc_fieldmap.dy 
        dz = tc_fieldmap.dz 
        scale = [1., dx, dy, dz, 
                dx * dy, dx * dz, dy * dz, 
                dx * dy * dz]

        phi = fieldmap.phi
        #print(phi.shape)

        ##########################################################################
        # dphi_dx = np.zeros_like(phi)
        # dphi_dy = np.zeros_like(phi)
        # dphi_dxdy = np.zeros_like(phi)
        # dphi_dx[1:-1,:] = 0.5*(phi[2:,:] - phi[:-2,:])
        # dphi_dy[:,1:-1] = 0.5*(phi[:,2:] - phi[:,:-2])
        # dphi_dxdy[1:-1,:] = 0.5*(dphi_dy[2:,:] - dphi_dy[:-2,:])
        
        # print("Setting derivatives: ")
        # kk=0.
        # for ix in range(nx):
        #     if (ix)/nx > kk:
        #         while (ix)/nx > kk:
        #             kk += 0.1
        #         print(f"{int(np.round(100*kk)):d}%..")
        #     for iy in range(ny):
        #         for iz in range(nz):
        #             index = 8 * ix + 8 * nx * iy + 8 * nx * ny * iz
        #             tc_fieldmap._phi_taylor[index+0] = phi[ix, iy, 0]
        #             tc_fieldmap._phi_taylor[index+1] = dphi_dx[ix, iy, 0]
        #             tc_fieldmap._phi_taylor[index+2] = dphi_dy[ix, iy, 0]
        #             tc_fieldmap._phi_taylor[index+3] = 0.
        #             tc_fieldmap._phi_taylor[index+4] = dphi_dxdy[ix, iy, 0]
        #             tc_fieldmap._phi_taylor[index+5] = 0.
        #             tc_fieldmap._phi_taylor[index+6] = 0.
        #             tc_fieldmap._phi_taylor[index+7] = 0.
        ##########################################################################

        ## Optimized version of above block ##########################################
        phi_slice = np.zeros([phi.shape[0], phi.shape[1], 8])
        for iz in range(nz):
            phi_slice[:,:,0] = phi[:,:, iz]
            phi_slice[1:-1,:,1] =  0.5*(phi[2:,:,iz] - phi[:-2,:,iz])
            phi_slice[:,1:-1,2] = 0.5*(phi[:,2:,iz] - phi[:,:-2,iz])
            phi_slice[1:-1,:,4] = 0.5*(phi_slice[2:,:,2] - phi_slice[:-2,:,2])

            flat_slice = phi_slice.transpose(1,0,2).flatten()
            len_slice = len(flat_slice)
            index_offset = 8 * nx * ny * iz
            tc_fieldmap._phi_taylor[index_offset:index_offset+len_slice] = flat_slice
        ##############################################################################

        self.xoinitialize(
                 _context=_context,
                 _buffer=_buffer,
                 _offset=_offset,
                 current=current,
                 length=length,
                 voltage=voltage,
                 fieldmap=tc_fieldmap)


srcs = []
srcs.append(_pkg_root.joinpath('headers/constants.h'))
srcs.append(_pkg_root.joinpath('fieldmaps/interpolated_src/tricubic_coefficients.h'))
srcs.append(_pkg_root.joinpath('fieldmaps/interpolated_src/cubic_interpolators.h'))
srcs.append(_pkg_root.joinpath('beam_elements/electronlens_src/electronlens_interpolated.h'))

ElectronLensInterpolated.XoStruct.extra_sources = srcs


# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

import xtrack as xt


def configure_orbit_dependent_parameters_for_bb(tracker, particle_on_co):

    """
    Assumes that the separation is saved in mean_x, mean_y for the 4D
    bb lenses.
    """

    temp_particles = particle_on_co.copy()
    for ii, ee in enumerate(tracker.line.elements):
        if ee.__class__.__name__ == 'BeamBeamBiGaussian2D':
              px_0 = temp_particles.px[0]
              py_0 = temp_particles.py[0]

              # Separation of 4D is so far set w.r.t. the closes orbit
              # (to be able to compare against sixtrack)
              # Here we set the right quantities (coordinates of the strong beam)
              ee.mean_x += temp_particles.x[0]
              ee.mean_y += temp_particles.y[0]

              ee.track(temp_particles)

              ee.d_px = temp_particles.px[0] - px_0
              ee.d_py = temp_particles.py[0] - py_0

              temp_particles.px -= ee.d_px
              temp_particles.py -= ee.d_py

        elif ee.__class__.__name__ == 'BeamBeamBiGaussian3D':
            ee.x_CO = temp_particles.x[0]
            ee.px_CO = temp_particles.px[0]
            ee.y_CO = temp_particles.y[0]
            ee.py_CO = temp_particles.py[0]
            ee.sigma_CO = temp_particles.zeta[0]
            ee.delta_CO = temp_particles.delta[0]

            ee.track(temp_particles)

            ee.Dx_sub = temp_particles.x[0] - ee.x_CO
            ee.Dpx_sub = temp_particles.px[0] - ee.px_CO
            ee.Dy_sub = temp_particles.y[0] - ee.y_CO
            ee.Dpy_sub = temp_particles.py[0] - ee.py_CO
            ee.Dsigma_sub = temp_particles.zeta[0] - ee.sigma_CO
            ee.Ddelta_sub = temp_particles.delta[0] - ee.delta_CO

            temp_particles.x[0] = ee.x_CO
            temp_particles.px[0] = ee.px_CO
            temp_particles.y[0] = ee.y_CO
            temp_particles.py[0] = ee.py_CO
            temp_particles.zeta[0] = ee.sigma_CO
            temp_particles.delta[0] = ee.delta_CO

        else:
            ee.track(temp_particles)



# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

from scipy.constants import e as qe
import xobjects as xo
import xtrack as xt

from ..general import _pkg_root

class BoostParameters(xo.Struct):
    sphi = xo.Float64
    cphi = xo.Float64
    tphi = xo.Float64
    salpha = xo.Float64
    calpha = xo.Float64

class Sigmas(xo.Struct):
    Sig_11 = xo.Float64
    Sig_12 = xo.Float64
    Sig_13 = xo.Float64
    Sig_14 = xo.Float64
    Sig_22 = xo.Float64
    Sig_23 = xo.Float64
    Sig_24 = xo.Float64
    Sig_33 = xo.Float64
    Sig_34 = xo.Float64
    Sig_44 = xo.Float64


class BeamBeamBiGaussian3D(xt.BeamElement):

    _xofields = {
        'q0': xo.Float64,
        'boost_parameters': BoostParameters,
        'Sigmas_0_star': Sigmas,
        'min_sigma_diff': xo.Float64,
        'threshold_singular': xo.Float64,
        'num_slices': xo.Int64,
        'delta_x': xo.Float64,
        'delta_y': xo.Float64,
        'x_CO': xo.Float64,
        'px_CO': xo.Float64,
        'y_CO': xo.Float64,
        'py_CO': xo.Float64,
        'sigma_CO': xo.Float64,
        'delta_CO': xo.Float64,
        'Dx_sub': xo.Float64,
        'Dpx_sub': xo.Float64,
        'Dy_sub': xo.Float64,
        'Dpy_sub': xo.Float64,
        'Dsigma_sub': xo.Float64,
        'Ddelta_sub': xo.Float64,
        'N_part_per_slice': xo.Float64[:],
        'x_slices_star': xo.Float64[:],
        'y_slices_star': xo.Float64[:],
        'sigma_slices_star': xo.Float64[:],
    }

    _input_param_names = [
        'phi',
        'alpha',
        'zeta_slices',
        'Sig_11_0',
        'Sig_12_0',
        'Sig_13_0',
        'Sig_14_0',
        'Sig_22_0',
        'Sig_23_0',
        'Sig_24_0',
        'Sig_33_0',
        'Sig_34_0',
        'Sig_44_0',
        ]

    def to_dict(self):

        dct = super().to_dict()
        for nn in self._input_param_names:
            dct[nn] = getattr(self, nn)
        # For compatibility with ducktrack:
        dct['x_bb_co'] = self.delta_x
        dct['y_bb_co'] = self.delta_y
        dct['x_co'] = self.x_CO
        dct['px_co'] = self.px_CO
        dct['y_co'] = self.y_CO
        dct['py_co'] = self.py_CO
        dct['zeta_co'] = self.sigma_CO
        dct['delta_co'] = self.delta_CO
        dct['d_x'] = self.Dx_sub
        dct['d_px'] = self.Dpx_sub
        dct['d_y'] = self.Dy_sub
        dct['d_py'] = self.Dpy_sub
        dct['d_zeta'] = self.Dsigma_sub
        dct['d_delta'] = self.Ddelta_sub
        dct['charge_slices'] = self.N_part_per_slice
        dct['sigma_11'] = self.Sig_11_0
        dct['sigma_12'] = self.Sig_12_0
        dct['sigma_13'] = self.Sig_13_0
        dct['sigma_14'] = self.Sig_14_0
        dct['sigma_22'] = self.Sig_22_0
        dct['sigma_23'] = self.Sig_23_0
        dct['sigma_24'] = self.Sig_24_0
        dct['sigma_33'] = self.Sig_33_0
        dct['sigma_34'] = self.Sig_34_0
        dct['sigma_44'] = self.Sig_44_0
        return dct


    def __init__(self, **kwargs):

        if 'old_interface' in kwargs:
            params=kwargs['old_interface']
            n_slices=len(params["charge_slices"])
            super().__init__(
                N_part_per_slice=n_slices,
                x_slices_star=n_slices,
                y_slices_star=n_slices,
                sigma_slices_star=n_slices)
            self._from_oldinterface(params)
        else:
            super().__init__(**kwargs)
            for nn in self._input_param_names:
                if nn in kwargs.keys():
                    setattr(self, nn, kwargs[nn])

    def _from_oldinterface(self, params):
        import ducktrack

        from scipy.constants import e as qe
        bb6d_data = ducktrack.BB6Ddata.BB6D_init(
                q_part=qe, # the ducktrack input has the charge
                phi=params["phi"],
                alpha=params["alpha"],
                delta_x=params["x_bb_co"],
                delta_y=params["y_bb_co"],
                N_part_per_slice=params["charge_slices"], #
                z_slices=params["zeta_slices"],
                Sig_11_0=params["sigma_11"],
                Sig_12_0=params["sigma_12"],
                Sig_13_0=params["sigma_13"],
                Sig_14_0=params["sigma_14"],
                Sig_22_0=params["sigma_22"],
                Sig_23_0=params["sigma_23"],
                Sig_24_0=params["sigma_24"],
                Sig_33_0=params["sigma_33"],
                Sig_34_0=params["sigma_34"],
                Sig_44_0=params["sigma_44"],
                x_CO=params["x_co"],
                px_CO=params["px_co"],
                y_CO=params["y_co"],
                py_CO=params["py_co"],
                sigma_CO=params["zeta_co"],
                delta_CO=params["delta_co"],
                min_sigma_diff=1e-10,
                threshold_singular=1e-28,
                Dx_sub=params["d_x"],
                Dpx_sub=params["d_px"],
                Dy_sub=params["d_y"],
                Dpy_sub=params["d_py"],
                Dsigma_sub=params["d_zeta"],
                Ddelta_sub=params["d_delta"],
                enabled=1,
            )
        assert(
            len(bb6d_data.N_part_per_slice) ==
            len(bb6d_data.x_slices_star) ==
            len(bb6d_data.y_slices_star) ==
            len(bb6d_data.sigma_slices_star))

        self.q0 = bb6d_data.q_part/qe, # ducktrack uses coulomb
        self.boost_parameters = {
                 'sphi': bb6d_data.parboost.sphi,
                 'cphi': bb6d_data.parboost.cphi,
                 'tphi': bb6d_data.parboost.tphi,
                 'salpha': bb6d_data.parboost.salpha,
                 'calpha': bb6d_data.parboost.calpha}
        self.Sigmas_0_star = {
                 'Sig_11': bb6d_data.Sigmas_0_star.Sig_11_0,
                 'Sig_12': bb6d_data.Sigmas_0_star.Sig_12_0,
                 'Sig_13': bb6d_data.Sigmas_0_star.Sig_13_0,
                 'Sig_14': bb6d_data.Sigmas_0_star.Sig_14_0,
                 'Sig_22': bb6d_data.Sigmas_0_star.Sig_22_0,
                 'Sig_23': bb6d_data.Sigmas_0_star.Sig_23_0,
                 'Sig_24': bb6d_data.Sigmas_0_star.Sig_24_0,
                 'Sig_33': bb6d_data.Sigmas_0_star.Sig_33_0,
                 'Sig_34': bb6d_data.Sigmas_0_star.Sig_34_0,
                 'Sig_44': bb6d_data.Sigmas_0_star.Sig_44_0}
        self.min_sigma_diff = bb6d_data.min_sigma_diff
        self.threshold_singular = bb6d_data.threshold_singular
        self.delta_x = bb6d_data.delta_x
        self.delta_y = bb6d_data.delta_y
        self.x_CO = bb6d_data.x_CO
        self.px_CO = bb6d_data.px_CO
        self.y_CO = bb6d_data.y_CO
        self.py_CO = bb6d_data.py_CO
        self.sigma_CO = bb6d_data.sigma_CO
        self.delta_CO = bb6d_data.delta_CO
        self.Dx_sub = bb6d_data.Dx_sub
        self.Dpx_sub = bb6d_data.Dpx_sub
        self.Dy_sub = bb6d_data.Dy_sub
        self.Dpy_sub = bb6d_data.Dpy_sub
        self.Dsigma_sub = bb6d_data.Dsigma_sub
        self.Ddelta_sub = bb6d_data.Ddelta_sub
        self.num_slices = len(bb6d_data.N_part_per_slice)
        self.N_part_per_slice = bb6d_data.N_part_per_slice
        self.x_slices_star = bb6d_data.x_slices_star
        self.y_slices_star = bb6d_data.y_slices_star
        self.sigma_slices_star = bb6d_data.sigma_slices_star

        # input_params
        self.phi = params['phi']
        self.alpha = params['alpha']
        self.zeta_slices = params['zeta_slices']
        self.Sig_11_0 = params["sigma_11"]
        self.Sig_12_0 = params["sigma_12"]
        self.Sig_13_0 = params["sigma_13"]
        self.Sig_14_0 = params["sigma_14"]
        self.Sig_22_0 = params["sigma_22"]
        self.Sig_23_0 = params["sigma_23"]
        self.Sig_24_0 = params["sigma_24"]
        self.Sig_33_0 = params["sigma_33"]
        self.Sig_34_0 = params["sigma_34"]
        self.Sig_44_0 = params["sigma_44"]

srcs = []
srcs.append(_pkg_root.joinpath('headers/constants.h'))
srcs.append(_pkg_root.joinpath('headers/sincos.h'))
srcs.append(_pkg_root.joinpath('headers/power_n.h'))
srcs.append(_pkg_root.joinpath('fieldmaps/bigaussian_src/complex_error_function.h'))
srcs.append('#define NOFIELDMAP') #TODO Remove this workaound
srcs.append(_pkg_root.joinpath('fieldmaps/bigaussian_src/bigaussian.h'))
srcs.append(_pkg_root.joinpath('beam_elements/beambeam_src/beambeam3d.h'))

BeamBeamBiGaussian3D.XoStruct.extra_sources = srcs



# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

import xobjects as xo
import xtrack as xt

import numpy as np
from numpy import sqrt, pi
from scipy.special import gamma

from ..general import _pkg_root

class LongitudinalProfileQGaussianData(xo.Struct):
    number_of_particles = xo.Float64
    _q_tol = xo.Float64
    _z0 = xo.Float64
    _sigma_z = xo.Float64
    _q_param = xo.Float64
    _cq_param = xo.Float64
    _beta_param = xo.Float64
    _sqrt_beta_param = xo.Float64
    _support_min = xo.Float64
    _support_max = xo.Float64

LongitudinalProfileQGaussianData.extra_sources = [
    _pkg_root.joinpath('longitudinal_profiles/qgaussian_src/qgaussian.h')
    ]
LongitudinalProfileQGaussianData.custom_kernels = {'line_density_qgauss':
        xo.Kernel(args=[xo.Arg(LongitudinalProfileQGaussianData, name='prof'),
                        xo.Arg(xo.Int64, name='n'),
                        xo.Arg(xo.Float64, pointer=True, name='z'),
                        xo.Arg(xo.Float64, pointer=True, name='res')],
                  n_threads='n')}




class LongitudinalProfileQGaussian(xo.dress(LongitudinalProfileQGaussianData)):

    @staticmethod
    def cq_from_q(q, q_tol):
        cq = sqrt(pi)
        if q >= (1 + q_tol):
            cq *= gamma((3 - q) / (2 * q - 2))
            cq /= sqrt((q - 1)) * gamma(1 / (q - 1))
        elif q <= (1 - q_tol):
            cq *= 2 * gamma(1 / (1 - q))
            cq /= (
                (3 - q)
                * sqrt(1 - q)
                * gamma((3 - q) / (2 - 2 * q))
            )
        return cq

    def __init__(self,
            _context=None,
            _buffer=None,
            _offset=None,
            number_of_particles=None,
            sigma_z=None,
            z0=0.,
            q_parameter=1.,
            z_min = -1e10,
            z_max = 1e10,
            q_tol=1e-6):


        assert number_of_particles is not None
        assert sigma_z is not None

        self.xoinitialize(_context=_context, _buffer=_buffer, _offset=_offset)

        self._z_min = z_min
        self._z_max = z_max
        self.number_of_particles = number_of_particles
        self.sigma_z = sigma_z
        self.z0 = z0
        self.q_tol = q_tol
        self.q_parameter = q_parameter

    def _recompute_beta_param(self):
        self._beta_param = 1./(self.sigma_z*self.sigma_z*(5.-3.*self.q_parameter))
        self._sqrt_beta_param = np.sqrt(self._beta_param)

    def _recompute_support(self):
        support_min = self._z_min
        support_max = self._z_max
        # Handle limited support
        if self.q_parameter < (1. - self.q_tol):
            rng = 1./sqrt(self.beta_param*(1-self.q_parameter))
            allowed_min = self.z0 - rng
            allowed_max = self.z0 + rng
            if support_min < allowed_min:
                support_min = allowed_min
            if support_max > allowed_max:
                support_max = allowed_max
        self._support_min = support_min
        self._support_max = support_max

    @property
    def sigma_z(self):
        return self._sigma_z

    @sigma_z.setter
    def sigma_z(self, value):
        self._sigma_z = value
        self._recompute_beta_param()
        self._recompute_support()

    @property
    def z0(self):
        return self._z0

    @z0.setter
    def z0(self, value):
        self._z0 = value
        self._recompute_support()

    @property
    def q_parameter(self):
        return self._q_param

    @q_parameter.setter
    def q_parameter(self, value):
        if value >= 5./3.:
            raise NotImplementedError
        self._q_param = value
        self._cq_param = self.__class__.cq_from_q(value, self.q_tol)
        self._recompute_beta_param()
        self._recompute_support()

    @property
    def q_tol(self):
        return self._q_tol

    @q_tol.setter
    def q_tol(self, value):
        self._q_tol = value
        self._recompute_support()

    @property
    def beta_param(self):
        return self._beta_param

    @property
    def z_min(self):
        return self._z_min

    @z_min.setter
    def z_min(self, value):
        self._z_min = value
        self._recompute_support()

    @property
    def z_max(self):
        return self._z_max

    @z_max.setter
    def z_max(self, value):
        self._z_max = value
        self._recompute_support()

    def line_density(self, z):
        context = self._buffer.context
        res = context.zeros(len(z), dtype=np.float64)

        if 'line_density_q_gauss' not in context.kernels.keys():
            self.compile_custom_kernels()

        context.kernels.line_density_qgauss(prof=self._xobject, n=len(z), z=z, res=res)

        return res




# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

import numpy as np

import xfields as xf
import xpart as xp
import xtrack as xt


def get_electroncloud_fieldmap_from_h5(
        filename, tau_max=None, buffer=None, ecloud_name="e-cloud"):
    assert buffer is not None
    import h5py
    ff = h5py.File(filename, "r")

    nx = len(ff["grid/xg"][()])
    ix1 = 0
    ix2 = nx

    ny = len(ff["grid/yg"][()])
    iy1 = 0
    iy2 = ny

    # Select a subset of slices (from -tau_max to +tau_max)
    if tau_max is None:
        nz = len(ff["grid/zg"][()])
        iz1 = 0
        iz2 = nz
    else:
        zg = ff["grid/zg"][()]
        nz = len(zg)
        # add one index to make sure range is included
        min_index = np.argmin(np.abs(zg + np.abs(tau_max))) - 1
        max_index = np.argmin(np.abs(zg - np.abs(tau_max))) + 1

        if min_index < 0 or max_index > nz:
            raise Exception(
                f"Range ({-np.abs(tau_max):.4f}, {np.abs(tau_max):.4f}) not in maximum range of z_grid: ({zg[0]:.4f},{zg[-1:]:.3f}) of file: {filename}")

        iz1 = min_index
        iz2 = max_index

    x_grid = ff["grid/xg"][ix1:ix2]
    y_grid = ff["grid/yg"][iy1:iy2]
    z_grid = ff["grid/zg"][iz1:iz2]

    mirror2D = ff["settings/symmetric2D"][()]
    # (in GB), 8 bytes per double-precision number
    memory_estimate = (ix2 - ix1) * (iy2 - iy1) * (iz2 - iz1) * 8 * 8 * 1.e-9
    print(f"Creating fieldmap... (Memory estimate = {memory_estimate:.2f} GB)")
    fieldmap = xf.TriCubicInterpolatedFieldMap(x_grid=x_grid, y_grid=y_grid, z_grid=z_grid,
                                               mirror_x=mirror2D, mirror_y=mirror2D, mirror_z=0, _buffer=buffer)
    print(f"Reading {ecloud_name}: ")
    kk = 0.
    scale = [1., fieldmap.dx, fieldmap.dy, fieldmap.dz,
             fieldmap.dx * fieldmap.dy, fieldmap.dx *
             fieldmap.dz, fieldmap.dy * fieldmap.dz,
             fieldmap.dx * fieldmap.dy * fieldmap.dz]

    ####### Optimized version of the loop in the block below. ################
    for iz in range(iz1, iz2):
        if (iz - iz1) / (iz2 - iz1) > kk:
            while (iz - iz1) / (iz2 - iz1) > kk:
                kk += 0.2
            print(f"{int(np.round(100*kk)):d}%..")
        phi_slice = ff[f"slices/slice{iz}/phi"][ix1:ix2,
                                                iy1:iy2, :].transpose(1, 0, 2)
        for ll in range(8):
            phi_slice[:, :, ll] *= scale[ll]
        index_offset = 8 * nx * ny * (iz - iz1)
        len_slice = phi_slice.shape[0] * \
            phi_slice.shape[1] * phi_slice.shape[2]
        fieldmap._phi_taylor[index_offset:index_offset +
                             len_slice] = phi_slice.flatten()
    ##########################################################################
    # for iz in range(iz1, iz2):
    #     if (iz-iz1)/nz > kk:
    #         while (iz-iz1)/nz > kk:
    #             kk += 0.1
    #         print(f"{int(np.round(100*kk)):d}%..")
    #     phi_slice = ff[f"slices/slice{iz}/phi"][ix1:ix2, iy1:iy2,:]
    #     for iy in range(ny):
    #         for ix in range(nx):
    #             for ll in range(8):
    #                 index = ll + 8 * ix + 8 * nx * iy + 8 * nx * ny * (iz - iz1)
    #                 fieldmap._phi_taylor[index] = phi_slice[ix, iy, ll] * scale[ll]
    ##########################################################################

    return fieldmap


def insert_electronclouds(eclouds, fieldmap=None, line=None):
    assert line is not None
    for name in eclouds.keys():
        s = eclouds[name]["s"]
        length = 0.
        line.insert_element(
            element=xf.ElectronCloud(
                length=length,
                fieldmap=fieldmap,
                _buffer=fieldmap._buffer),
            name=name,
            at_s=s)


def config_electronclouds(line, twiss=None, ecloud_info=None, shift_to_closed_orbit=False,
                          subtract_dipolar_kicks=False, fieldmaps=None, ecloud_strength=1.):
    assert twiss is not None
    assert ecloud_info is not None
    if subtract_dipolar_kicks:
        dipolar_kicks = {}
        assert fieldmaps is not None
        for key in fieldmaps.keys():
            fieldmap = fieldmaps[key]
            dipolar_kicks[key] = electroncloud_dipolar_kicks_of_fieldmap(
                fieldmap=fieldmap, p0c=line.particle_ref.p0c)

    length_factor = ecloud_strength / \
        (line.particle_ref.p0c * line.particle_ref.beta0)
    part = twiss["particle_on_co"].copy()
    for ii, el_name in enumerate(twiss["name"]):
        if 'ecloud' in el_name:
            # naming format is "ecloud.ecloud_type.sector.index_in_sector",
            # e.g.  ecloud.mb.78.38
            ecloud_type = el_name.split(".")[1]
            length = ecloud_info[ecloud_type][el_name]["length"] * \
                length_factor
            line.elements[ii].length = length
            assert el_name == line.element_names[ii]

            if shift_to_closed_orbit:
                line.elements[ii].x_shift = twiss["x"][ii]
                line.elements[ii].y_shift = twiss["y"][ii]
                part.delta = twiss["delta"][ii]
                part.zeta = twiss["zeta"][ii]
                line.elements[ii].tau_shift = part.zeta[0] / \
                    (part.beta0[0] * part.rvv[0])

            if subtract_dipolar_kicks:
                line.elements[ii].dipolar_px_kick = dipolar_kicks[ecloud_type][0] * length
                line.elements[ii].dipolar_py_kick = dipolar_kicks[ecloud_type][1] * length
                line.elements[ii].dipolar_ptau_kick = dipolar_kicks[ecloud_type][2] * length


def electroncloud_dipolar_kicks_of_fieldmap(fieldmap=None, p0c=None):

    assert p0c is not None
    assert fieldmap is not None

    part = xp.Particles(p0c=p0c)
    ecloud = xf.ElectronCloud(
        length=1,
        fieldmap=fieldmap,
        _buffer=fieldmap._buffer)
    ecloud.track(part)
    px = part.px[0]
    py = part.py[0]
    ptau = part.ptau[0]
    return [px, py, ptau]


def full_electroncloud_setup(line=None, ecloud_info=None, filenames=None, context=None,
                             tau_max=None, subtract_dipolar_kicks=True, shift_to_closed_orbit=True):

    buffer = context.new_buffer()
    fieldmaps = {
        ecloud_type: get_electroncloud_fieldmap_from_h5(
            filename=filename,
            buffer=buffer,
            tau_max=tau_max,
            ecloud_name=ecloud_type) for (
            ecloud_type,
            filename) in filenames.items()}

    for ecloud_type, fieldmap in fieldmaps.items():
        print(f"Inserting \"{ecloud_type}\" electron clouds...")
        insert_electronclouds(
            ecloud_info[ecloud_type],
            fieldmap=fieldmap,
            line=line)

    tracker = xt.Tracker(_context=context, line=line, _buffer=buffer)
    twiss_without_ecloud = tracker.twiss()
    config_electronclouds(
        line,
        twiss=twiss_without_ecloud,
        ecloud_info=ecloud_info,
        subtract_dipolar_kicks=subtract_dipolar_kicks,
        shift_to_closed_orbit=shift_to_closed_orbit,
        fieldmaps=fieldmaps,
        ecloud_strength=1)
    twiss_with_ecloud = tracker.twiss()

    return tracker, twiss_without_ecloud, twiss_with_ecloud



# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

import numpy as np

import xobjects as xo
import xtrack as xt
from xobjects import context_default

def mean_and_std(a, weights=None):
    if weights is None:
        mean = a.sum()/len(a)
        std = np.sqrt(((a-mean)**2).sum() / len(a))
    else:
        assert len(weights) == len(a)
        tot = weights.sum()
        mean = (a*weights).sum() / tot
        std = np.sqrt(((a-mean)**2 * weights).sum() / tot)

    return float(mean), float(std)

class BiGaussianFieldMap(xo.DressedStruct):

    '''
    Builds a field-map object that provides the fields generated by a
    normalized 2D Gaussian distribution (Bassetti-Erskine formula).

    Args:
        context (xobjects context): identifies the :doc:`context <contexts>`
            on which the computation is executed.
        mean_x (float64): Horizontal position (in meters) of the Gaussian
            distribution. It can be updated after the object creation.
            Default is ``0.``.
        mean_y (float64): Vertical position (in meters) of the Gaussian
            distribution. It can be updated after the object creation.
            Default is ``0.``.
        sigma_x (float64): Horizontal r.m.s. size (in meters) of the Gaussian
            distribution. It can be updated after the object creation.
            Default is ``None``.
        sigma_y (float64): Vertical r.m.s. size (in meters) of tthe Gaussian
            distribution. It can be updated after the object creation.
            Default is ``None``.
        min_sigma_diff (float64): Difference between sigma_x and sigma_y (in
            meters) below which round distribution is assumed.
        updatable (bool): If ``True`` the field map can be updated after
            creation. Default is ``True``.
    Returns:
        (BiGaussianFieldMap): Field map object.
    '''

    _xofields={
            'mean_x': xo.Float64,
            'mean_y': xo.Float64,
            'sigma_x': xo.Float64,
            'sigma_y': xo.Float64,
            'min_sigma_diff': xo.Float64,
            '_updatable': xo.Int64
        }

    def __init__(self,
                 _context=None,
                 _buffer=None,
                 _offset=None,
                 _xobject=None,
                 mean_x=0., mean_y=0.,
                 sigma_x=None, sigma_y=None,
                 min_sigma_diff=1e-10,
                 updatable=True):

        if _xobject is not None:
            self.xoinitialize(_xobject=_xobject)
        else:
            self.xoinitialize(
                     _context=_context,
                     _buffer=_buffer,
                     _offset=_offset)

            self.updatable = updatable
            self.mean_x = mean_x
            self.mean_y = mean_y
            self.sigma_x = sigma_x
            self.sigma_y = sigma_y
            self.min_sigma_diff=min_sigma_diff

    @property
    def updatable(self):
        return bool(self._updatable)

    @updatable.setter
    def updatable(self, value):
        self._updatable=np.int64(value)

#    def get_values_at_points(self,
#            x, y,
#            return_rho=False,
#            return_phi=False,
#            return_dphi_dx=True,
#            return_dphi_dy=True,
#            ):
#
#        """
#        Returns the derivatives ot the electric potential at the points specified by x, y.
#        The output can be customized (see below).
#
#        Args:
#            x (float64 array): Horizontal coordinates at which the field is evaluated.
#            y (float64 array): Vertical coordinates at which the field is evaluated.
#            return_rho (bool): If ``True``, the charge density at the given points is
#                returned. Default is ``False``.
#            return_phi (bool): If ``True``, the potential at the given points is returned.
#            return_dphi_dx (bool): If ``True``, the horizontal derivative of the potential
#                at the given points is returned. Default is ``True``.
#            return_dphi_dy: If ``True``, the vertical derivative of the potential
#                at the given points is returned. Default is ``True``.
#        Returns:
#            (tuple of float64 array): The required quantities at the provided points.
#        """
#        if self.sigma_x is None:
#            raise ValueError('sigma_x must be set')
#        if self.sigma_y is None:
#            raise ValueError('sigma_y must be set')
#
#        assert len(x) == len(y)
#        tobereturned = []
#
#        if return_rho:
#            raise notimplementederror('not yet implemented :-(')
#        if return_phi:
#            raise notimplementederror('not yet implemented :-(')
#
#        if return_dphi_dx or return_dphi_dy:
#            Ex = self.context.zeros(x.shape, dtype=np.float64)
#            Ey = self.context.zeros(x.shape, dtype=np.float64)
#            self.context.kernels.get_Ex_Ey_Gx_Gy_gauss(
#                n_points=len(x),
#                x_ptr=x-self.mean_x,
#                y_ptr=y-self.mean_y,
#                sigma_x=self.sigma_x,
#                sigma_y=self.sigma_y,
#                min_sigma_diff=self.min_sigma_diff,
#                skip_Gs=1,
#                Ex_ptr=Ex,
#                Ey_ptr=Ey,
#                Gx_ptr=Ex, # untouchd when skip_Gs is zero
#                Gy_ptr=Ex, # untouchd when skip_Gs is zero
#                )
#            if return_dphi_dx:
#                tobereturned.append(-Ex)
#            if return_dphi_dy:
#                tobereturned.append(-Ey)
#
#        return tobereturned

    def update_from_particles(self, x_p, y_p, z_p, ncharges_p, q0_coulomb,
                reset=True, update_phi=True, solver=None, force=False):

        if not force:
            self._assert_updatable()

        raise NotImplementedError()

    def update_rho(self, rho, reset):
        raise ValueError('rho cannot be directly updated'
                         'for BiGaussianFieldMap')

    def update_phi(self, phi, reset=True, force=False):
        raise ValueError('phi cannot be directly updated'
                         'for BiGaussianFieldMap')

    def update_phi_from_rho(self, solver=None):
        raise ValueError('phi cannot be directly updated'
                         'for BiGaussianFieldMap')

    def generate_solver(self, solver):
        raise ValueError('solver cannot be generated'
                         'for BiGaussianFieldMap')

BiGaussianFieldMapData = BiGaussianFieldMap.XoStruct



# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

from abc import ABC, abstractmethod


class Solver(ABC):

    @abstractmethod
    def __init__(self, context=None, **kwargs):
        pass

    @abstractmethod
    def solve(self, rho):
        return phi





# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

from .longitudinal_profiles import LongitudinalProfileCoasting
from .longitudinal_profiles import LongitudinalProfileQGaussian

from .fieldmaps import TriLinearInterpolatedFieldMap
from .fieldmaps import TriCubicInterpolatedFieldMap
from .fieldmaps import BiGaussianFieldMap, mean_and_std

from .solvers.fftsolvers import FFTSolver3D

from .beam_elements.spacecharge import SpaceCharge3D, SpaceChargeBiGaussian
from .beam_elements.beambeam import BeamBeamBiGaussian2D
from .beam_elements.beambeam3d import BeamBeamBiGaussian3D
from .beam_elements.electroncloud import ElectronCloud
from .beam_elements.electronlens_interpolated import ElectronLensInterpolated

from .general import _pkg_root
from .config_tools import replace_spacecharge_with_quasi_frozen
from .config_tools import replace_spacecharge_with_PIC
from .config_tools import configure_orbit_dependent_parameters_for_bb
from .config_tools import install_spacecharge_frozen
from .config_tools import full_electroncloud_setup


import xtrack as _xt

element_classes = tuple(
    v for v in globals().values()
    if isinstance(v, type) and issubclass(v, _xt.BeamElement)
)
del _xt



# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #


class LongitudinalProfileCoasting(object):


    def __init__(self, context=None, beam_line_density=None):

        if context is None:
            context = ContextDefault()

        assert beam_line_density is not None

        self.beam_line_density = beam_line_density

    def line_density(self, z):
        return self.beam_line_density



# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

import numpy as np
from scipy.constants import e as qe
from scipy.constants import c as clight

from xfields import BiGaussianFieldMap, mean_and_std
from xfields import TriLinearInterpolatedFieldMap
from ..longitudinal_profiles import LongitudinalProfileQGaussianData
from ..longitudinal_profiles import LongitudinalProfileQGaussian
from ..fieldmaps import BiGaussianFieldMapData
from ..fieldmaps import TriLinearInterpolatedFieldMapData
from ..general import _pkg_root

import xobjects as xo
import xtrack as xt


class SpaceCharge3D(xt.BeamElement):

    """
    Simulates the effect of space charge on a bunch.

    Args:
        context (XfContext): identifies the :doc:`context <contexts>`
            on which the computation is executed.
        update_on_track (bool): If ``True`` the beam field map is update
            at each interaction. If ``False`` the initial field map is
            used at each interaction (frozen model). The default is
            ``True``.
        length (float): the length of the space-charge interaction in
            meters.
        apply_z_kick (bool): If ``True``, the longitudinal kick on the
            particles is applied.
        x_range (tuple): Horizontal extent (in meters) of the
            computing grid.
        y_range (tuple): Vertical extent (in meters) of the
            computing grid.
        z_range (tuple): Longitudina extent  (in meters) of
            the computing grid.
        nx (int): Number of cells in the horizontal direction.
        ny (int): Number of cells in the vertical direction.
        nz (int): Number of cells in the vertical direction.
        dx (float): Horizontal cell size in meters. It can be
            provided alternatively to ``nx``.
        dy (float): Vertical cell size in meters. It can be
            provided alternatively to ``ny``.
        dz (float): Longitudinal cell size in meters.It can be
            provided alternatively to ``nz``.
        x_grid (np.ndarray): Equispaced array with the horizontal grid points
            (cell centers).
            It can be provided alternatively to ``x_range``, ``dx``/``nx``.
        y_grid (np.ndarray): Equispaced array with the horizontal grid points
            (cell centers).
            It can be provided alternatively to ``y_range``, ``dy``/``ny``.
        z_grid (np.ndarray): Equispaced array with the horizontal grid points
            (cell centers).
            It can be provided alternatively to ``z_range``, ``dz``/``nz``.
        rho (np.ndarray): initial charge density at the grid points in
            Coulomb/m^3.
        phi (np.ndarray): initial electric potential at the grid points in
            Volts. If not provided the ``phi`` is calculated from ``rho``
            using the Poisson solver (if available).
        solver (str or solver object): Defines the Poisson solver to be used
            to compute phi from rho. Accepted values are ``FFTSolver3D`` and
            ``FFTSolver2p5D``. A Xfields solver object can also be provided.
            In case ``update_on_track``is ``False`` and ``phi`` is provided
            by the user, this argument can be omitted.
        gamma0 (float): Relativistic gamma factor of the beam. This is required
            only if the solver is ``FFTSolver3D``.
    Returns:
        (SpaceCharge3D): A space-charge 3D beam element.
    """
    _xofields = {
        'fieldmap': xo.Ref(TriLinearInterpolatedFieldMapData),
        'length': xo.Float64,
        }

    def copy(self, _context=None, _buffer=None, _offset=None):
        if _buffer is not self._buffer:
            raise NotImplementedError
        return SpaceCharge3D(_context=_context,
                _buffer=_buffer, _offset=_offset,
                update_on_track=self.update_on_track,
                length=self.length,
                apply_z_kick=self.apply_z_kick,
                fieldmap=self.fieldmap)

    def __init__(self,
                 _context=None,
                 _buffer=None,
                 _offset=None,
                 update_on_track=True,
                 length=None,
                 apply_z_kick=True,
                 fieldmap=None,
                 x_range=None, y_range=None, z_range=None,
                 nx=None, ny=None, nz=None,
                 dx=None, dy=None, dz=None,
                 x_grid=None, y_grid=None, z_grid=None,
                 rho=None, phi=None,
                 solver=None,
                 gamma0=None,
                 fftplan=None):

        self.update_on_track = update_on_track
        self.apply_z_kick = apply_z_kick

        if solver=='FFTSolver3D':
            assert gamma0 is not None, ('To use FFTSolver3D '
                                        'gamma0 must be provided')

        if gamma0 is not None:
            if not np.isscalar(gamma0):
                raise ValueError('gamma0 needs to be a scalar')
            scale_coordinates_in_solver=(1.,1., float(gamma0))
        else:
            scale_coordinates_in_solver=(1.,1.,1.)

        if _buffer is not None:
            _context = _buffer.context
        if _context is None:
            _context = xo.context_default

        if fieldmap is None:
            # I build the fieldmap on a temporary buffer
            temp_buff = _context.new_buffer()
            fieldmap = TriLinearInterpolatedFieldMap(
                        _buffer=temp_buff,
                        rho=rho, phi=phi,
                        x_grid=z_grid, y_grid=y_grid, z_grid=z_grid,
                        x_range=x_range, y_range=y_range, z_range=z_range,
                        dx=dx, dy=dy, dz=dz,
                        nx=nx, ny=ny, nz=nz,
                        solver=solver,
                        scale_coordinates_in_solver=scale_coordinates_in_solver,
                        updatable=update_on_track,
                        fftplan=fftplan)

        self.xoinitialize(
                 _context=_context,
                 _buffer=_buffer,
                 _offset=_offset,
                 fieldmap=fieldmap,
                 length=length)

        # temp_buff is deallocate here

    @property
    def iscollective(self):
        return self.update_on_track


    def track(self, particles):

        """
        Computes and applies the space-charge forces for the provided set of
        particles.

        Args:
            particles (Particles Object): Particles to be tracked.
        """

        if self.update_on_track:
            self.fieldmap.update_from_particles(
                particles=particles)

        # call C tracking kernel
        super().track(particles)


srcs = []
srcs.append(_pkg_root.joinpath('headers/constants.h'))
srcs.append(_pkg_root.joinpath('fieldmaps/interpolated_src/linear_interpolators.h'))
srcs.append(_pkg_root.joinpath('beam_elements/spacecharge_src/spacecharge3d.h'))

SpaceCharge3D.XoStruct.extra_sources = srcs



class SpaceChargeBiGaussian(xt.BeamElement):

    _xofields = {
        'longitudinal_profile': LongitudinalProfileQGaussianData, # TODO: Will become unionref
        'fieldmap': BiGaussianFieldMapData,
        'length': xo.Float64,
        }

    def to_dict(self):
        dct = super().to_dict()
        # To be loaded by ducktrack:
        dct['number_of_particles'] = self.longitudinal_profile.number_of_particles
        dct['bunchlength_rms'] = self.longitudinal_profile.sigma_z
        dct['sigma_x'] = self.fieldmap.sigma_x
        dct['sigma_y'] = self.fieldmap.sigma_y
        dct['x_co'] = self.fieldmap.mean_x
        dct['y_co'] = self.fieldmap.mean_y
        return dct

    def __init__(self,
                 _context=None,
                 _buffer=None,
                 _offset=None,
                 _xobject=None,
                 update_on_track=False,
                 length=None,
                 apply_z_kick=False,
                 longitudinal_profile=None,
                 mean_x=0.,
                 mean_y=0.,
                 sigma_x=None,
                 sigma_y=None,
                 fieldmap=None,
                 min_sigma_diff=1e-10,
                 **kwargs # to avoid issues when building form dict
                 ):

        if _xobject is not None:
            self.xoinitialize(
                     _context=_context,
                     _buffer=_buffer,
                     _offset=_offset,
                     _xobject=_xobject)
        else:
            self.xoinitialize(
                     _context=_context,
                     _buffer=_buffer,
                     _offset=_offset)

            if apply_z_kick:
                raise NotImplementedError

            assert longitudinal_profile is not None, (
                'Longitudinal profile must be provided')

            self.length = length
            self.longitudinal_profile = longitudinal_profile
            self.apply_z_kick = apply_z_kick
            self._init_update_on_track(update_on_track)

            if fieldmap is None:
                self.fieldmap = BiGaussianFieldMap(
                         _context=self._buffer.context,
                         mean_x=mean_x,
                         mean_y=mean_y,
                         sigma_x=sigma_x,
                         sigma_y=sigma_y,
                         min_sigma_diff=min_sigma_diff,
                         updatable=True)
            else:
                self.fieldmap=fieldmap

        self.iscollective = None # Inferred from _update_flag

    def track(self, particles):

        if self._update_flag:
            mean_x, sigma_x = mean_and_std(
                    particles.x, weights=particles.weight)
            mean_y, sigma_y = mean_and_std(
                    particles.y, weights=particles.weight)
            if self.update_mean_x_on_track:
                self.mean_x = mean_x
            if self.update_mean_y_on_track:
                self.mean_y = mean_y
            if self.update_sigma_x_on_track:
                self.sigma_x = sigma_x
            if self.update_sigma_y_on_track:
                self.sigma_y = sigma_y

        super().track(particles)


    def _init_update_on_track(self, update_on_track):
        self.update_mean_x_on_track = False
        self.update_mean_y_on_track = False
        self.update_sigma_x_on_track = False
        self.update_sigma_y_on_track = False
        if update_on_track == True:
            self.update_mean_x_on_track = True
            self.update_mean_y_on_track = True
            self.update_sigma_x_on_track = True
            self.update_sigma_y_on_track = True
        elif update_on_track == False:
            pass
        else:
            for nn in update_on_track:
                assert nn in ['mean_x', 'mean_y',
                              'sigma_x', 'sigma_y']
                setattr(self, f'update_{nn}_on_track', True)

    @property
    def _update_flag(self):
        return (self.update_mean_x_on_track or
                self.update_mean_y_on_track or
                self.update_sigma_x_on_track or
                self.update_sigma_y_on_track)

    @property
    def iscollective(self):
        if self._iscollective is not None:
            return self._iscollective
        else:
            return self._update_flag

    @iscollective.setter
    def iscollective(self, value):
        self._iscollective = value

    @property
    def mean_x(self):
        return self.fieldmap.mean_x

    @ mean_x.setter
    def mean_x(self, value):
        self.fieldmap.mean_x = value

    @property
    def mean_y(self):
        return self.fieldmap.mean_y

    @ mean_y.setter
    def mean_y(self, value):
        self.fieldmap.mean_y = value

    @property
    def sigma_x(self):
        return self.fieldmap.sigma_x

    @ sigma_x.setter
    def sigma_x(self, value):
        self.fieldmap.sigma_x = value

    @property
    def sigma_y(self):
        return self.fieldmap.sigma_y

    @ sigma_y.setter
    def sigma_y(self, value):
        self.fieldmap.sigma_y = value


srcs = []
srcs.append(_pkg_root.joinpath('headers/constants.h'))
srcs.append(_pkg_root.joinpath('headers/sincos.h'))
srcs.append(_pkg_root.joinpath('headers/power_n.h'))
srcs.append(_pkg_root.joinpath('fieldmaps/bigaussian_src/complex_error_function.h'))
srcs.append(_pkg_root.joinpath('fieldmaps/bigaussian_src/bigaussian.h'))
srcs.append(_pkg_root.joinpath('longitudinal_profiles/qgaussian_src/qgaussian.h'))
srcs.append(_pkg_root.joinpath('beam_elements/spacecharge_src/spacechargebigaussian.h'))

SpaceChargeBiGaussian.XoStruct.extra_sources = srcs



# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

import numpy as np

from xpart import Particles

def generate_particles_object(
                            n_macroparticles,
                            bunch_intensity,
                            sigma_x,
                            sigma_y,
                            sigma_z,
                            p0c,
                            mass,
                            n_probes,
                            r_max_probes,
                            z_probes,
                            theta_probes):

    x_part = sigma_x * np.random.normal(size=(n_macroparticles,))
    y_part = sigma_y * np.random.normal(size=(n_macroparticles,))
    z_part = sigma_z * np.random.normal(size=(n_macroparticles,))
    weights_part = 0*x_part + bunch_intensity/n_macroparticles

    # insert probes
    r_probes= np.linspace(-r_max_probes, r_max_probes, n_probes)
    x_probes = r_probes * np.cos(theta_probes)
    y_probes = r_probes * np.sin(theta_probes)
    z_probes = 0 * x_probes +z_probes

    x_part = np.concatenate([x_probes, x_part])
    y_part = np.concatenate([y_probes, y_part])
    z_part = np.concatenate([z_probes, z_part])
    weights_part = np.concatenate([0*x_probes, weights_part])


    particles = Particles(
            p0c=p0c,
            mass0 = mass,
            x=x_part,
            y=y_part,
            zeta=z_part,
            )
    particles.weight = weights_part

    return particles, r_probes, x_probes, y_probes, z_probes




# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

import numpy as np

import xobjects as xo
import xpart as xp
import xtrack as xt

from ..solvers.fftsolvers import FFTSolver3D, FFTSolver2p5D
from ..general import _pkg_root

class TriLinearInterpolatedFieldMapData(xo.Struct):
    x_min = xo.Float64
    y_min = xo.Float64
    z_min = xo.Float64
    nx = xo.Int64
    ny = xo.Int64
    nz = xo.Int64
    dx = xo.Float64
    dy = xo.Float64
    dz = xo.Float64
    rho = xo.Float64[:]
    phi = xo.Float64[:]
    dphi_dx = xo.Float64[:]
    dphi_dy = xo.Float64[:]
    dphi_dz = xo.Float64[:]

TriLinearInterpolatedFieldMapData.extra_sources = [
    _pkg_root.joinpath('headers/constants.h'),
    _pkg_root.joinpath('fieldmaps/interpolated_src/central_diff.h'),
    _pkg_root.joinpath('fieldmaps/interpolated_src/linear_interpolators.h'),
    _pkg_root.joinpath('fieldmaps/interpolated_src/charge_deposition.h'),
    ]

TriLinearInterpolatedFieldMapData.custom_kernels = {
    'central_diff': xo.Kernel(
        args=[
            xo.Arg(xo.Int32,   pointer=False, name='nelem'),
            xo.Arg(xo.Int32,   pointer=False, name='row_size'),
            xo.Arg(xo.Int32,   pointer=False, name='stride_in_dbl'),
            xo.Arg(xo.Float64, pointer=False, name='factor'),
            xo.Arg(xo.Int8,    pointer=True,  name='matrix_buffer'),
            xo.Arg(xo.Int64,   pointer=False, name='matrix_offset'),
            xo.Arg(xo.Int8,    pointer=True,  name='res_buffer'),
            xo.Arg(xo.Int64,   pointer=False, name='res_offset'),
            ],
        n_threads='nelem'
        ),
    'p2m_rectmesh3d_xparticles': xo.Kernel(
        args=[
            xo.Arg(xo.Int32,   pointer=False, name='nparticles'),
            xo.Arg(xp.Particles.XoStruct, pointer=False, name='particles'),
            xo.Arg(xo.Float64, pointer=False, name='x0'),
            xo.Arg(xo.Float64, pointer=False, name='y0'),
            xo.Arg(xo.Float64, pointer=False, name='z0'),
            xo.Arg(xo.Float64, pointer=False, name='dx'),
            xo.Arg(xo.Float64, pointer=False, name='dy'),
            xo.Arg(xo.Float64, pointer=False, name='dz'),
            xo.Arg(xo.Int32,   pointer=False, name='nx'),
            xo.Arg(xo.Int32,   pointer=False, name='ny'),
            xo.Arg(xo.Int32,   pointer=False, name='nz'),
            xo.Arg(xo.Int8,    pointer=True,  name='grid1d_buffer'),
            xo.Arg(xo.Int64,   pointer=False, name='grid1d_offset'),
            ],
        n_threads='nparticles'
        ),
    'p2m_rectmesh3d': xo.Kernel(
        args=[
            xo.Arg(xo.Int32,   pointer=False, name='nparticles'),
            xo.Arg(xo.Float64, pointer=True, name='x'),
            xo.Arg(xo.Float64, pointer=True, name='y'),
            xo.Arg(xo.Float64, pointer=True, name='z'),
            xo.Arg(xo.Float64, pointer=True, name='part_weights'),
            xo.Arg(xo.Int64,   pointer=True, name='part_state'),
            xo.Arg(xo.Float64, pointer=False, name='x0'),
            xo.Arg(xo.Float64, pointer=False, name='y0'),
            xo.Arg(xo.Float64, pointer=False, name='z0'),
            xo.Arg(xo.Float64, pointer=False, name='dx'),
            xo.Arg(xo.Float64, pointer=False, name='dy'),
            xo.Arg(xo.Float64, pointer=False, name='dz'),
            xo.Arg(xo.Int32,   pointer=False, name='nx'),
            xo.Arg(xo.Int32,   pointer=False, name='ny'),
            xo.Arg(xo.Int32,   pointer=False, name='nz'),
            xo.Arg(xo.Int8,    pointer=True,  name='grid1d_buffer'),
            xo.Arg(xo.Int64,   pointer=False, name='grid1d_offset'),
            ],
        n_threads='nparticles'
        ),
    'TriLinearInterpolatedFieldMap_interpolate_3d_map_vector': xo.Kernel(
        args=[
            xo.Arg(TriLinearInterpolatedFieldMapData, pointer=False, name='fmap'),
            xo.Arg(xo.Int64,   pointer=False, name='n_points'),
            xo.Arg(xo.Float64, pointer=True,  name='x'),
            xo.Arg(xo.Float64, pointer=True,  name='y'),
            xo.Arg(xo.Float64, pointer=True,  name='z'),
            xo.Arg(xo.Int64,   pointer=False, name='n_quantities'),
            xo.Arg(xo.Int8,    pointer=True,  name='buffer_mesh_quantities'),
            xo.Arg(xo.Int64,   pointer=True,  name='offsets_mesh_quantities'),
            xo.Arg(xo.Float64, pointer=True,  name='particles_quantities'),
            ],
        n_threads='n_points'
        ),
    }

# I add undescores in front of the names so that I can define custom properties
rename_trilinear = {ff.name:'_'+ff.name for ff
                in TriLinearInterpolatedFieldMapData._fields}
class TriLinearInterpolatedFieldMap(xo.dress(TriLinearInterpolatedFieldMapData,
                                             rename=rename_trilinear)):

    """
    Builds a linear interpolator for a 3D field map. The map can be updated
    using the Parcle In Cell method.

    Args:
        context (xobjects context): identifies the :doc:`context <contexts>`
            on which the computation is executed.
        x_range (tuple): Horizontal extent (in meters) of the
            computing grid.
        y_range (tuple): Vertical extent (in meters) of the
            computing grid.
        z_range (tuple): Longitudina extent  (in meters) of
            the computing grid.
        nx (int): Number of cells in the horizontal direction.
        ny (int): Number of cells in the vertical direction.
        nz (int): Number of cells in the vertical direction.
        dx (float): Horizontal cell size in meters. It can be
            provided alternatively to ``nx``.
        dy (float): Vertical cell size in meters. It can be
            provided alternatively to ``ny``.
        dz (float): Longitudinal cell size in meters.It can be
            provided alternatively to ``nz``.
        x_grid (np.ndarray): Equispaced array with the horizontal grid points
            (cell centers).
            It can be provided alternatively to ``x_range``, ``dx``/``nx``.
        y_grid (np.ndarray): Equispaced array with the horizontal grid points
            (cell centers).
            It can be provided alternatively to ``y_range``, ``dy``/``ny``.
        z_grid (np.ndarray): Equispaced array with the horizontal grid points
            (cell centers).
            It can be provided alternatively to ``z_range``, ``dz``/``nz``.
        rho (np.ndarray): initial charge density at the grid points in
            Coulomb/m^3.
        phi (np.ndarray): initial electric potential at the grid points in
            Volts. If not provided the ``phi`` is calculated from ``rho``
            using the Poisson solver (if available).
        solver (str or solver object): Defines the Poisson solver to be used
            to compute phi from rho. Accepted values are ``FFTSolver3D`` and
            ``FFTSolver2p5D``. A Xfields solver object can also be provided.
            In case ``update_on_track``is ``False`` and ``phi`` is provided
            by the user, this argument can be omitted.
        scale_coordinates_in_solver (tuple): Three coefficients used to rescale
            the grid coordinates in the definition of the solver. The default is
            (1.,1.,1.).
        updatable (bool): If ``True`` the field map can be updated after
            creation. Default is ``True``.
    Returns:
        (TriLinearInterpolatedFieldMap): Interpolator object.
    """

    def __init__(self,
                 _context=None,
                 _buffer=None,
                 _offset=None,
                 x_range=None, y_range=None, z_range=None,
                 nx=None, ny=None, nz=None,
                 dx=None, dy=None, dz=None,
                 x_grid=None, y_grid=None, z_grid=None,
                 rho=None, phi=None,
                 solver=None,
                 scale_coordinates_in_solver=(1.,1.,1.),
                 updatable=True,
                 fftplan=None
                 ):


        self.updatable = updatable
        self.scale_coordinates_in_solver = scale_coordinates_in_solver

        self._x_grid = _configure_grid('x', x_grid, dx, x_range, nx)
        self._y_grid = _configure_grid('y', y_grid, dy, y_range, ny)
        self._z_grid = _configure_grid('z', z_grid, dz, z_range, nz)

        nelem = self.nx*self.ny*self.nz
        self.xoinitialize(
                 _context=_context,
                 _buffer=_buffer,
                 _offset=_offset,
                 x_min = self._x_grid[0],
                 y_min = self._y_grid[0],
                 z_min = self._z_grid[0],
                 nx = self.nx,
                 ny = self.ny,
                 nz = self.nz,
                 dx = self.dx,
                 dy = self.dy,
                 dz = self.dz,
                 rho = nelem,
                 phi = nelem,
                 dphi_dx = nelem,
                 dphi_dy = nelem,
                 dphi_dz = nelem)

        self.compile_custom_kernels(only_if_needed=True)

        if isinstance(solver, str):
            self.solver = self.generate_solver(solver, fftplan)
        else:
            #TODO: consistency check to be added
            self.solver = solver

        # Set rho
        if rho is not None:
            self.update_rho(rho, force=True)

        # Set phi
        if phi is not None:
            self.update_phi(phi, force=True)
        else:
            if solver is not None and rho is not None:
                self.update_phi_from_rho()

    def _assert_updatable(self):
        assert self.updatable, 'This FieldMap is not updatable!'

    #@profile
    def get_values_at_points(self,
            x, y, z,
            return_rho=True,
            return_phi=True,
            return_dphi_dx=True,
            return_dphi_dy=True,
            return_dphi_dz=True):

        """
        Returns the charge density, the field potential and its derivatives
        at the points specified by x, y, z. The output can be customized (see below).
        Zeros are returned for points outside the grid.

        Args:
            x (float64 array): Horizontal coordinates at which the field is evaluated.
            y (float64 array): Vertical coordinates at which the field is evaluated.
            z (float64 array): Longitudinal coordinates at which the field is evaluated.
            return_rho (bool): If ``True``, the charge density at the given points is
                returned.
            return_phi (bool): If ``True``, the potential at the given points is returned.
            return_dphi_dx (bool): If ``True``, the horizontal derivative of the potential
                at the given points is returned.
            return_dphi_dy: If ``True``, the vertical derivative of the potential
                at the given points is returned.
            return_dphi_dz: If ``True``, the longitudinal derivative of the potential
                at the given points is returned.
        Returns:
            (tuple of float64 array): The required quantities at the provided points.
        """

        assert len(x) == len(y) == len(z)

        pos_in_buffer_of_maps_to_interp = []
        if return_rho:
            pos_in_buffer_of_maps_to_interp.append(
                    self._xobject.rho._offset + self._xobject.rho._data_offset)
        if return_phi:
            pos_in_buffer_of_maps_to_interp.append(
                    self._xobject.phi._offset + self._xobject.phi._data_offset)
        if return_dphi_dx:
            pos_in_buffer_of_maps_to_interp.append(
                    self._xobject.dphi_dx._offset + self._xobject.dphi_dx._data_offset)
        if return_dphi_dy:
            pos_in_buffer_of_maps_to_interp.append(
                    self._xobject.dphi_dy._offset + self._xobject.dphi_dy._data_offset)
        if return_dphi_dz:
            pos_in_buffer_of_maps_to_interp.append(
                    self._xobject.dphi_dz._offset + self._xobject.dphi_dz._data_offset)

        context = self._buffer.context

        pos_in_buffer_of_maps_to_interp = context.nparray_to_context_array(
                        np.array(pos_in_buffer_of_maps_to_interp, dtype=np.int64))
        nmaps_to_interp = len(pos_in_buffer_of_maps_to_interp)
        buffer_out = context.zeros(
                shape=(nmaps_to_interp * len(x),), dtype=np.float64)
        if nmaps_to_interp > 0:
            context.kernels.TriLinearInterpolatedFieldMap_interpolate_3d_map_vector(
                    fmap=self._xobject,
                    n_points=len(x),
                    x=x, y=y, z=z,
                    n_quantities=nmaps_to_interp,
                    buffer_mesh_quantities=self._buffer.buffer,
                    offsets_mesh_quantities=pos_in_buffer_of_maps_to_interp,
                    particles_quantities=buffer_out)

        # Split buffer 
        particles_quantities = [buffer_out[ii*len(x):(ii+1)*len(x)]
                                        for ii in range(nmaps_to_interp)]

        return particles_quantities

    #@profile
    def update_from_particles(self,
                        particles=None,
                        x_p=None, y_p=None, z_p=None,
                        ncharges_p=None, state_p=None, q0_coulomb=None,
                        reset=True, update_phi=True, solver=None, force=False):

        """
        Updates the charge density at the grid using a given set of particles,
        which can be provided by a particles object or by individual arrays.
        The potential can be optionally updated accordingly.

        Args:
            particles (xtrack.Particles): xtrack particle object.
            x_p (float64 array): Horizontal coordinates of the macroparticles.
            y_p (float64 array): Vertical coordinates of the macroparticles.
            z_p (float64 array): Longitudinal coordinates of the macroparticles.
            ncharges_p (float64 array): Number of reference charges in the
                macroparticles.
            state_p (int64, array): particle state (>0 active, lost otherwise)
            q0_coulomb (float64): Reference charge in Coulomb.
            reset (bool): If ``True`` the stored charge density is overwritten
                with the provided one. If ``False`` the provided charge density
                is added to the stored one. The default is ``True``.
            update_phi (bool): If ``True`` the stored potential is recalculated
                from the stored charge density.
            solver (Solver object): solver object to be used to solve Poisson's
                equation (compute phi from rho). If ``None`` is provided the solver
                attached to the fieldmap is used (if any). The default is ``None``.
            force (bool): If ``True`` the potential is updated even if the
                map is declared as not updateable. The default is ``False``.
        """

        if not force:
            self._assert_updatable()

        if reset:
            self.rho[:,:,:] = 0.

        context = self._buffer.context

        if particles is None:
            assert (len(x_p) == len(y_p) == len(z_p) == len(ncharges_p))
            if state_p is None:
                state_p = context.zeros(shape=x_p.shape, dtype=np.int64) + 1
            else:
                assert len(state_p) == len(x_p)

            context.kernels.p2m_rectmesh3d(
                    nparticles=len(x_p),
                    x=x_p, y=y_p, z=z_p,
                    part_weights=q0_coulomb*ncharges_p,
                    part_state=state_p,
                    x0=self.x_grid[0], y0=self.y_grid[0], z0=self.z_grid[0],
                    dx=self.dx, dy=self.dy, dz=self.dz,
                    nx=self.nx, ny=self.ny, nz=self.nz,
                    grid1d_buffer=self._xobject.rho._buffer.buffer,
                    grid1d_offset=self._xobject.rho._offset
                                 +self._xobject.rho._data_offset)
        else:
            assert (x_p is None and y_p is None and z_p is None
                    and ncharges_p is None and state_p is None)
            context.kernels.p2m_rectmesh3d_xparticles(
                    nparticles=particles._capacity,
                    particles=particles,
                    x0=self.x_grid[0], y0=self.y_grid[0], z0=self.z_grid[0],
                    dx=self.dx, dy=self.dy, dz=self.dz,
                    nx=self.nx, ny=self.ny, nz=self.nz,
                    grid1d_buffer=self._xobject.rho._buffer.buffer,
                    grid1d_offset=self._xobject.rho._offset
                                 +self._xobject.rho._data_offset)

        if update_phi:
            self.update_phi_from_rho(solver=solver)

    def update_rho(self, rho, reset=True, force=False):
        """
        Updates the charge density on the grid.

        Args:
            rho (float64 array): Charge density at the grid points in C/m^3.
            reset (bool): If ``True`` the stored charge density is overwritten
                with the provided one. If ``False`` the provided charge density
                is added to the stored one. The default is ``True``.
            force (bool): If ``True`` the charge density is updated even if the
                map is declared as not updateable. The default is ``False``.
        """

        if not force:
            self._assert_updatable()

        if reset:
            self.rho[:,:,:] = rho
        else:
            raise ValueError('Not implemented!')

    #@profile
    def update_phi(self, phi, reset=True, force=False):

        """
        Updates the potential on the grid. The stored derivatives are also
        updated.

        Args:
            rho (float64 array): Potential at the grid points.
            reset (bool): If ``True`` the stored potential is overwritten
                with the provided one. If ``False`` the provided potential
                is added to the stored one. The default is ``True``.
            force (bool): If ``True`` the potential is updated even if the
                map is declared as not updateable. The default is ``False``.
        """

        if not force:
            self._assert_updatable()

        if reset:
            self.phi.T[:,:,:] = phi.T
        else:
            raise ValueError('Not implemented!')

        context = self._buffer.context

        # Compute gradient
        context.kernels.central_diff(
                nelem = self.phi.size,
                row_size = self.nx,
                stride_in_dbl = self.phi.strides[0]/8,
                factor = 1/(2*self.dx),
                matrix_buffer = self._xobject.phi._buffer.buffer,
                matrix_offset = (self._xobject.phi._offset
                               + self._xobject.phi._data_offset),
                res_buffer = self._xobject.dphi_dx._buffer.buffer,
                res_offset = (self._xobject.dphi_dx._offset
                            + self._xobject.dphi_dx._data_offset))
        context.kernels.central_diff(
                nelem = self.phi.size,
                row_size = self.ny,
                stride_in_dbl = self.phi.strides[1]/8,
                factor = 1/(2*self.dy),
                matrix_buffer = self._xobject.phi._buffer.buffer,
                matrix_offset = (self._xobject.phi._offset
                               + self._xobject.phi._data_offset),
                res_buffer = self._xobject.dphi_dy._buffer.buffer,
                res_offset = (self._xobject.dphi_dy._offset
                            + self._xobject.dphi_dy._data_offset))
        context.kernels.central_diff(
                nelem = self.phi.size,
                row_size = self.nz,
                stride_in_dbl = self.phi.strides[2]/8,
                factor = 1/(2*self.dz),
                matrix_buffer = self._xobject.phi._buffer.buffer,
                matrix_offset = (self._xobject.phi._offset
                               + self._xobject.phi._data_offset),
                res_buffer = self._xobject.dphi_dz._buffer.buffer,
                res_offset = (self._xobject.dphi_dz._offset
                            + self._xobject.dphi_dz._data_offset))

    #@profile
    def update_phi_from_rho(self, solver=None):

        """
        Updates the potential on the grid (phi) from the charge density on the
        grid (phi). It requires a :doc:`Poisson solver object <solvers>`. If
        none is provided the one attached to the fieldmap is used (if any).

        Args:
            solver (Solver object): solver object to be used to solve Poisson's
                equation. If ``None`` is provided the solver attached to the fieldmap
                is used (if any). The default is ``None``.
        """

        self._assert_updatable()

        if solver is None:
            if hasattr(self, 'solver'):
                solver = self.solver
            else:
                raise ValueError('I have no solver to compute phi!')

        new_phi = solver.solve(self.rho)
        self.update_phi(new_phi)

    def generate_solver(self, solver, fftplan):

        """
        Generates a Poisson solver associated to the defined grid.

        Args:
            solver (str): Defines the Poisson solver to be used
            to compute phi from rho. Accepted values are ``FFTSolver3D`` and
            ``FFTSolver2p5D``.
        Returns:
            (Solver): Solver object associated to the defined grid.
        """

        scale_dx, scale_dy, scale_dz = self.scale_coordinates_in_solver

        if solver == 'FFTSolver3D':
            solver = FFTSolver3D(
                    dx=self.dx*scale_dx,
                    dy=self.dy*scale_dy,
                    dz=self.dz*scale_dz,
                    nx=self.nx, ny=self.ny, nz=self.nz,
                    context=self._buffer.context,
                    fftplan=fftplan)
        elif solver == 'FFTSolver2p5D':
            solver = FFTSolver2p5D(
                    dx=self.dx*scale_dx,
                    dy=self.dy*scale_dy,
                    dz=self.dz*scale_dz,
                    nx=self.nx, ny=self.ny, nz=self.nz,
                    context=self._buffer.context,
                    fftplan=fftplan)
        else:
            raise ValueError(f'solver name {solver} not recognized')

        return solver

    @property
    def x_grid(self):
        """
        Array with the horizontal grid points (cell centers).
        """
        return self._x_grid

    @property
    def y_grid(self):
        """
        Array with the vertical grid points (cell centers).
        """
        return self._y_grid

    @property
    def z_grid(self):
        """
        Array with the longitudinal grid points (cell centers).
        """
        return self._z_grid

    @property
    def nx(self):
        """
        Number of cells in the horizontal direction.
        """
        return len(self.x_grid)

    @property
    def ny(self):
        """
        Number of cells in the vertical direction.
        """
        return len(self.y_grid)

    @property
    def nz(self):
        """
        Number of cells in the longitudinal direction.
        """
        return len(self.z_grid)

    @property
    def dx(self):
        """
        Horizontal cell size in meters.
        """
        return self.x_grid[1] - self.x_grid[0]

    @property
    def dy(self):
        """
        Vertical cell size in meters.
        """
        return self.y_grid[1] - self.y_grid[0]

    @property
    def dz(self):
        """
        Longitudinal cell size in meters.
        """
        return self.z_grid[1] - self.z_grid[0]

    # TODO: these reshapes can be avoided by allocating 3d arrays directly in the xobject
    @property
    def rho(self):
        return self._rho.reshape(
                (self.nx, self.ny, self.nz), order='F')

    @property
    def phi(self):
        """
        Electric potential at the grid points in Volts.
        """
        return self._phi.reshape(
                (self.nx, self.ny, self.nz), order='F')

    @property
    def dphi_dx(self):
        return self._dphi_dx.reshape(
                (self.nx, self.ny, self.nz), order='F')

    @property
    def dphi_dy(self):
        return self._dphi_dy.reshape(
                (self.nx, self.ny, self.nz), order='F')

    @property
    def dphi_dz(self):
        return self._dphi_dz.reshape(
                (self.nx, self.ny, self.nz), order='F')



def _configure_grid(vname, v_grid, dv, v_range, nv):

    # Check input consistency
    if v_grid is not None:
        assert dv is None, (f'd{vname} cannot be given '
                            f'if {vname}_grid is provided ')
        assert nv is None, (f'n{vname} cannot be given '
                            f'if {vname}_grid is provided ')
        assert v_range is None, (f'{vname}_range cannot be given '
                                 f'if {vname}_grid is provided')
        ddd = np.diff(v_grid)
        assert np.allclose(ddd,ddd[0]), (f'{vname}_grid must be '
                                          'unifirmly spaced')
    else:
        assert v_range is not None, (f'{vname}_grid or {vname}_range '
                                     f'must be provided')
        assert len(v_range)==2, (f'{vname}_range must be in the form '
                                 f'({vname}_min, {vname}_max)')
        if dv is not None:
            assert nv is None, (f'n{vname} cannot be given '
                                    f'if d{vname} is provided ')
            v_grid = np.arange(v_range[0], v_range[1]+0.1*dv, dv)
        else:
            assert nv is not None, (f'n{vname} must be given '
                                    f'if d{vname} is not provided ')
            v_grid = np.linspace(v_range[0], v_range[1], nv)

    return v_grid




# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

from pathlib import Path
_pkg_root = Path(__file__).parent.absolute()



# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

import numpy as np
from scipy.constants import e as qe

from ..fieldmaps.bigaussian import BiGaussianFieldMap, BiGaussianFieldMapData
from ..general import _pkg_root

import xobjects as xo
import xtrack as xt


class BeamBeamBiGaussian2D(xt.BeamElement):
    """
    Simulates the effect of beam-beam on a bunch.

    Args:
        context (xobjects context): identifies the :doc:`context <contexts>`
            on which the computation is executed.
        n_particles (float64): Number of particles in the colliding bunch.
        q0 (float64): Number of particles in the colliding bunch.
        beta0 (float64): Relativistic beta of the colliding bunch.
        mean_x (float64): Horizontal position (in meters) of the colliding
            bunch. It can be updated after the object creation.
            Default is ``0.``.
        mean_y (float64): Vertical position (in meters) of the Gaussian
            distribution. It can be updated after the object creation.
            Default is ``0.``.
        sigma_x (float64): Horizontal r.m.s. size (in meters) of the colliding
            bunch. It can be updated after the object creation.
            Default is ``None``.
        sigma_y (float64): Vertical r.m.s. size (in meters) of the colliding
            bunch. It can be updated after the object creation.
            Default is ``None``.
    Returns:
        (BeamBeamBiGaussian2D): A beam-beam element.
    """
    _xofields={
        'n_particles': xo.Float64,
        'q0': xo.Float64,
        'beta0': xo.Float64,
        'fieldmap': BiGaussianFieldMapData,
        'd_px': xo.Float64,
        'd_py': xo.Float64,
        }

    def to_dict(self):
        dct = super().to_dict()
        dct['charge'] = self.q0*self.n_particles
        dct['sigma_x'] = self.sigma_x
        dct['sigma_y'] = self.sigma_y
        dct['beta_r'] = self.beta0
        dct['x_bb'] = self.mean_x
        dct['y_bb'] = self.mean_y
        return dct

    def __init__(self,
            _context=None,
            _buffer=None,
            _offset=None,
            _xobject=None,
            n_particles=None,
            q0=None,
            beta0=None,
            mean_x=0.,
            mean_y=0.,
            sigma_x=None,
            sigma_y=None,
            d_px=0.,
            d_py=0.,
            min_sigma_diff=1e-28,
            fieldmap=None,
            **kwargs # TODO: to be removed, needed to avoid problems in from_dict
            ):

        if _xobject is not None:

            self.xoinitialize(
                     _context=_context,
                     _buffer=_buffer,
                     _offset=_offset,
                     _xobject=_xobject)
        else:

            self.xoinitialize(
                     _context=_context,
                     _buffer=_buffer,
                     _offset=_offset)

            if not np.isscalar(beta0):
                raise ValueError('beta0 needs to be a scalar')

            self.n_particles = n_particles
            self.q0 = q0
            self.beta0 = beta0
            self.d_px = d_px
            self.d_py = d_py

            if fieldmap is None:
                fieldmap = BiGaussianFieldMap(
                         _context=_context,
                         _buffer=_buffer,
                         _offset=_offset,
                         mean_x=mean_x,
                         mean_y=mean_y,
                         sigma_x=sigma_x,
                         sigma_y=sigma_y,
                         min_sigma_diff=min_sigma_diff,
                         updatable=True)

            self.fieldmap=fieldmap

    def update(self, **kwargs):
        for kk in kwargs.keys():
            if not hasattr(self, kk):
                raise NameError(f'Unknown parameter: {kk}')
            setattr(self, kk, kwargs[kk])

    @property
    def mean_x(self):
        return self.fieldmap.mean_x

    @ mean_x.setter
    def mean_x(self, value):
        self.fieldmap.mean_x = value

    @property
    def mean_y(self):
        return self.fieldmap.mean_y

    @ mean_y.setter
    def mean_y(self, value):
        self.fieldmap.mean_y = value

    @property
    def sigma_x(self):
        return self.fieldmap.sigma_x

    @ sigma_x.setter
    def sigma_x(self, value):
        self.fieldmap.sigma_x = value

    @property
    def sigma_y(self):
        return self.fieldmap.sigma_y

    @ sigma_y.setter
    def sigma_y(self, value):
        self.fieldmap.sigma_y = value

srcs = []
srcs.append(_pkg_root.joinpath('headers/constants.h'))
srcs.append(_pkg_root.joinpath('headers/sincos.h'))
srcs.append(_pkg_root.joinpath('headers/power_n.h'))
srcs.append(_pkg_root.joinpath('fieldmaps/bigaussian_src/complex_error_function.h'))
srcs.append(_pkg_root.joinpath('fieldmaps/bigaussian_src/bigaussian.h'))
srcs.append(_pkg_root.joinpath('beam_elements/beambeam_src/beambeam.h'))

BeamBeamBiGaussian2D.XoStruct.extra_sources = srcs



# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

from .spacecharge_config_tools import *
from .beambeam_config_tools import configure_orbit_dependent_parameters_for_bb
from .electroncloud_config_tools import *


# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

from .interpolated import TriLinearInterpolatedFieldMap
from .interpolated import TriLinearInterpolatedFieldMapData
from .tricubicinterpolated import TriCubicInterpolatedFieldMap
from .tricubicinterpolated import TriCubicInterpolatedFieldMapData
from .bigaussian import BiGaussianFieldMap, mean_and_std
from .bigaussian import BiGaussianFieldMapData



# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

from .coasting import LongitudinalProfileCoasting
from .qgaussian import LongitudinalProfileQGaussian
from .qgaussian import LongitudinalProfileQGaussianData



# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

import numpy as np
import pandas as pd

from ..beam_elements.spacecharge import SpaceChargeBiGaussian
from ..beam_elements.spacecharge import SpaceCharge3D

import xpart as xp
import xtrack as xt
from xtrack.line import _is_thick, _is_drift

def install_spacecharge_frozen(line, particle_ref, longitudinal_profile,
                               nemitt_x, nemitt_y, sigma_z,
                               num_spacecharge_interactions,
                               tol_spacecharge_position,
                               s_spacecharge=None):

    tracker_no_sc = xt.Tracker(line=line.copy())

    # Make a matched bunch just to get the matched momentum spread
    bunch = xp.generate_matched_gaussian_bunch(
             num_particles=int(2e6), total_intensity_particles=1.,
             nemitt_x=nemitt_x, nemitt_y=nemitt_y, sigma_z=sigma_z,
             particle_ref=particle_ref, tracker=tracker_no_sc)
    delta_rms = np.std(bunch.delta)

    # Generate spacecharge positions
    if s_spacecharge is None:
        s_spacecharge = np.linspace(0, line.get_length(),
                                    num_spacecharge_interactions+1)[:-1]

    # Create spacecharge elements (dummy)
    sc_elements = []
    sc_names = []
    for ii, ss in enumerate(s_spacecharge):
        sc_elements.append(SpaceChargeBiGaussian(
            length=-9999,
            apply_z_kick=False,
            longitudinal_profile=longitudinal_profile,
            mean_x=0.,
            mean_y=0.,
            sigma_x=1.,
            sigma_y=1.))
        sc_names.append(f'spacecharge_{ii}')

        #TODO Replace loop with single insert_element when available in xtrack
        line.insert_element(name=sc_names[-1], element=sc_elements[-1],
                            at_s=ss, s_tol=tol_spacecharge_position)


    actual_s_spch = line.get_s_position(sc_names)

    sc_lengths = 0*s_spacecharge
    sc_lengths[:-1] = np.diff(actual_s_spch)
    sc_lengths[-1] = line.get_length() - np.sum(sc_lengths[:-1]) 

    # Twiss at spacecharge
    line_sc_off = line.filter_elements(exclude_types_starting_with='SpaceCh')
    tracker_sc_off = xt.Tracker(line=line_sc_off,
            element_classes=tracker_no_sc.element_classes,
            track_kernel=tracker_no_sc.track_kernel)
    tw_at_sc = tracker_sc_off.twiss(particle_ref=particle_ref, at_elements=sc_names)

    # Configure lenses
    for ii, sc in enumerate(sc_elements):
        sc.mean_x = tw_at_sc['x'][ii]
        sc.mean_y = tw_at_sc['y'][ii]
        sc.sigma_x = np.sqrt(tw_at_sc['betx'][ii]*nemitt_x
                               /particle_ref.beta0/particle_ref.gamma0
                             + (tw_at_sc['dx'][ii]*delta_rms)**2)
        sc.sigma_y = np.sqrt(tw_at_sc['bety'][ii]*nemitt_y
                               /particle_ref.beta0/particle_ref.gamma0
                             + (tw_at_sc['dy'][ii]*delta_rms)**2)
        sc.length = sc_lengths[ii]


def replace_spacecharge_with_quasi_frozen(
                        line, _buffer,
                        update_mean_x_on_track=True,
                        update_mean_y_on_track=True,
                        update_sigma_x_on_track=True,
                        update_sigma_y_on_track=True):

    spch_elements = []
    for ii, ee in enumerate(line.elements):
        if isinstance(ee, SpaceChargeBiGaussian):
            ee._move_to(_buffer=_buffer)
            ee.update_mean_x_on_track = update_mean_x_on_track
            ee.update_mean_y_on_track = update_mean_y_on_track
            ee.update_sigma_x_on_track = update_sigma_x_on_track
            ee.update_sigma_y_on_track = update_sigma_y_on_track
            ee.iscollective = True
            spch_elements.append(ee)

    return spch_elements


class PICCollection:

    def __init__(self,
                 nx_grid,
                 ny_grid,
                 nz_grid,
                 x_lim_min,
                 x_lim_max,
                 y_lim_min,
                 y_lim_max,
                 z_range,
                 n_lims_x,
                 n_lims_y,
                 solver='FFTSolver2p5D',
                 apply_z_kick=False,
                 _context=None,
                 _buffer=None,
                     ):

        self._context = _context
        self._buffer = _buffer

        self.nx_grid = nx_grid
        self.ny_grid = ny_grid
        self.nz_grid = nz_grid

        self.z_range = z_range
        self.solver = solver
        self.apply_z_kick = apply_z_kick

        self.x_lims = np.linspace(x_lim_min, x_lim_max, n_lims_x)
        self.y_lims = np.linspace(y_lim_min, y_lim_max, n_lims_y)

        self._existing_pics = {}
        self._fftplan = None


    def get_pic(self, x_lim, y_lim):

        assert x_lim < self.x_lims[-1]
        assert x_lim > self.x_lims[0]
        assert y_lim < self.y_lims[-1]
        assert y_lim > self.y_lims[0]

        ix = np.argmin(np.abs(x_lim - self.x_lims))
        iy = np.argmin(np.abs(y_lim - self.y_lims))

        if (ix, iy) not in self._existing_pics.keys():
            print(f'Creating PIC ({ix}, {iy})')
            xlim_pic = self.x_lims[ix]
            ylim_pic = self.y_lims[iy]
            new_pic = SpaceCharge3D(
                _context=self._context,
                _buffer=self._buffer,
                length=0.,
                apply_z_kick=self.apply_z_kick,
                x_range=(-xlim_pic, xlim_pic),
                y_range=(-ylim_pic, ylim_pic),
                z_range=self.z_range,
                nx=self.nx_grid, ny=self.ny_grid, nz=self.nz_grid,
                solver=self.solver,
                fftplan=self._fftplan)
            new_pic._buffer.grow(10*1024**2) # Add 10 MB for sc copies
            if self._fftplan is None:
                self._fftplan = new_pic.fieldmap.solver.fftplan
            self._existing_pics[ix, iy] = new_pic

        return self._existing_pics[ix, iy]


def replace_spacecharge_with_PIC(
        line,
        n_sigmas_range_pic_x, n_sigmas_range_pic_y,
        nx_grid, ny_grid, nz_grid, n_lims_x, n_lims_y, z_range,
        _context=None,
        _buffer=None):

    all_sc_elems = []
    name_sc_elems = []
    all_sigma_x = []
    all_sigma_y = []
    for ii, ee in enumerate(line.elements):
        if isinstance(ee, SpaceChargeBiGaussian):
            all_sc_elems.append(ee)
            name_sc_elems.append(line.element_names[ii])
            all_sigma_x.append(ee.sigma_x)
            all_sigma_y.append(ee.sigma_y)

    x_lim_min = np.min(all_sigma_x) * (n_sigmas_range_pic_x - 0.5)
    x_lim_max = np.max(all_sigma_x) * (n_sigmas_range_pic_x + 0.5)
    y_lim_min = np.min(all_sigma_y) * (n_sigmas_range_pic_y - 0.5)
    y_lim_max = np.max(all_sigma_y) * (n_sigmas_range_pic_y + 0.5)

    pic_collection = PICCollection(
        _context=_context,
        _buffer=_buffer,
        nx_grid=nx_grid, ny_grid=ny_grid, nz_grid=nz_grid,
        x_lim_min=x_lim_min, x_lim_max=x_lim_max, n_lims_x=n_lims_x,
        y_lim_min=y_lim_min, y_lim_max=y_lim_max, n_lims_y=n_lims_y,
        z_range=z_range)

    all_pics = []
    for nn, ee in zip(name_sc_elems, all_sc_elems):
        xlim = n_sigmas_range_pic_x*ee.sigma_x
        ylim = n_sigmas_range_pic_y*ee.sigma_y
        base_sc = pic_collection.get_pic(xlim, ylim)
        sc = base_sc.copy(_buffer=base_sc._buffer)
        sc.length = ee.length
        line.element_dict[nn] = sc
        all_pics.append(sc)

    return pic_collection, all_pics



# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

from ..fieldmaps import TriCubicInterpolatedFieldMapData
from ..general import _pkg_root

import xobjects as xo
import xtrack as xt


class ElectronCloud(xt.BeamElement):

    """
    Simulates the effect of an electron cloud on a bunch.

    Args:
        context (XfContext): identifies the :doc:`context <contexts>`
            on which the computation is executed.
        x_shift (float): shifts the x coordinate. Should be set equal to
            the closed orbit if the fieldmap is defined with respect to 
            the closed orbit. Measured in meters.
        y_shift (float): shifts the y coordinate. Should be set equal to
            the closed orbit if the fieldmap is defined with respect to 
            the closed orbit. Measured in meters.
        tau_shift (float): shifts the tau coordinate. Should be set equal 
            to the closed orbit if the fieldmap is defined with respect to 
            the closed orbit. Measured in meters. (tau = s/beta_0 - c t)
        dipolar_px_kick (float): subtracts a constant value from the kick to px.
            Should be set equal to the field map's kick at the closed orbit to 
            remove closed orbit distortion effects.
        dipolar_py_kick (float): subtracts a constant value from the kick to py.
            Should be set equal to the field map's kick at the closed orbit to 
            remove closed orbit distortion effects.
        dipolar_ptau_kick (float): subtracts a constant value from the kick to 
            ptau. Should be set equal to the field map's kick at the closed 
            orbit to remove closed orbit distortion effects.
        length (float): the length of the electron-cloud interaction in
            meters.
        apply_z_kick (bool): If ``True``, the longitudinal kick on the
            particles is applied. The default is ``True``.
        fieldmap (xfields.TriCubicInterpolatedFieldMap): Field map of the 
            electron cloud forces.
    Returns:
        (ElectronCloud): An electron cloud beam element.
    """

    _xofields = {
        'x_shift': xo.Float64,
        'y_shift': xo.Float64,
        'tau_shift': xo.Float64,
        'dipolar_px_kick': xo.Float64,
        'dipolar_py_kick': xo.Float64,
        'dipolar_ptau_kick': xo.Float64,
        'length': xo.Float64,
        #'fieldmap': TriCubicInterpolatedFieldMapData,
        'fieldmap': xo.Ref(TriCubicInterpolatedFieldMapData),
        }

    def __init__(self,
                 _context=None,
                 _buffer=None,
                 _offset=None,
                 x_shift=0.,
                 y_shift=0.,
                 tau_shift=0.,
                 dipolar_px_kick=0.,
                 dipolar_py_kick=0.,
                 dipolar_ptau_kick=0.,
                 length=None,
                 apply_z_kick=True,
                 fieldmap=None,
                 ):

        # To be implemented if false
        self.apply_z_kick = apply_z_kick
        if self.apply_z_kick is False:
            raise NotImplementedError

        if _buffer is not None:
            _context = _buffer.context
        if _context is None:
            _context = xo.context_default

        assert fieldmap is not None

        self.xoinitialize(
                 _context=_context,
                 _buffer=_buffer,
                 _offset=_offset,
                 x_shift=x_shift,
                 y_shift=y_shift,
                 tau_shift=tau_shift,
                 dipolar_px_kick=dipolar_px_kick,
                 dipolar_py_kick=dipolar_py_kick,
                 dipolar_ptau_kick=dipolar_ptau_kick,
                 length=length,
                 fieldmap=fieldmap)



srcs = []
srcs.append(_pkg_root.joinpath('fieldmaps/interpolated_src/tricubic_coefficients.h'))
srcs.append(_pkg_root.joinpath('fieldmaps/interpolated_src/cubic_interpolators.h'))
srcs.append(_pkg_root.joinpath('beam_elements/electroncloud_src/electroncloud.h'))

ElectronCloud.XoStruct.extra_sources = srcs



# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

from setuptools import setup, find_packages, Extension

#######################################
# Prepare list of compiled extensions #
#######################################

extensions = []


#########
# Setup #
#########

setup(
    name='xfields',
    version='0.6.1',
    description='Field Maps and Particle In Cell',
    url='https://github.com/xsuite/xfields',
    author='Giovanni Iadarola',
    packages=find_packages(),
    ext_modules = extensions,
    include_package_data=True,
    install_requires=[
        'numpy>=1.0',
        'scipy',
        'pandas',
        'xobjects>=0.0.4',
        'xtrack>=0.0.1',
        ]
    )



# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

import numpy as np
from scipy.constants import epsilon_0
from numpy import pi

from .base import Solver

from xobjects import context_default

class FFTSolver2D(Solver):

    def solve(self, rho):
        pass

class FFTSolver3D(Solver):

    '''
    Creates a Poisson solver object that solves the full 3D Poisson
    equation using the FFT method (free space).

    Args:
        nx (int): Number of cells in the horizontal direction.
        ny (int): Number of cells in the vertical direction.
        nz (int): Number of cells in the vertical direction.
        dx (float): Horizontal cell size in meters.
        dy (float): Vertical cell size in meters.
        dz (float): Longitudinal cell size in meters.
        context (XfContext): identifies the :doc:`context <contexts>`
            on which the computation is executed.
    Returns:
        (FFTSolver3D): Poisson solver object.
    '''

    def __init__(self, dx, dy, dz, nx, ny, nz, context=None, fftplan=None):

        if context is None:
            context = context_default

        self.context = context

        # Prepare arrays
        workspace_dev = context.nparray_to_context_array(
                    np.zeros((2*nx, 2*ny, 2*nz), dtype=np.complex128, order='F'))


        # Build grid for primitive function
        xg_F = np.arange(0, nx+2) * dx - dx/2
        yg_F = np.arange(0, ny+2) * dy - dy/2
        zg_F = np.arange(0, nz+2) * dz - dz/2
        XX_F, YY_F, ZZ_F = np.meshgrid(xg_F, yg_F, zg_F, indexing='ij')

        # Compute primitive
        F_temp = primitive_func_3d(XX_F, YY_F, ZZ_F)

        # Integrated Green Function (I will transform inplace)
        gint_rep= np.zeros((2*nx, 2*ny, 2*nz), dtype=np.complex128, order='F')
        gint_rep[:nx+1, :ny+1, :nz+1] = (F_temp[ 1:,  1:,  1:]
                                       - F_temp[:-1,  1:,  1:]
                                       - F_temp[ 1:, :-1,  1:]
                                       + F_temp[:-1, :-1,  1:]
                                       - F_temp[ 1:,  1:, :-1]
                                       + F_temp[:-1,  1:, :-1]
                                       + F_temp[ 1:, :-1, :-1]
                                       - F_temp[:-1, :-1, :-1])

        # Replicate
        # To define how to make the replicas I have a look at:
        # np.abs(np.fft.fftfreq(10))*10
        # = [0., 1., 2., 3., 4., 5., 4., 3., 2., 1.]
        gint_rep[nx+1:, :ny+1, :nz+1] = gint_rep[nx-1:0:-1, :ny+1,     :nz+1    ]
        gint_rep[:nx+1, ny+1:, :nz+1] = gint_rep[:nx+1,     ny-1:0:-1, :nz+1    ]
        gint_rep[nx+1:, ny+1:, :nz+1] = gint_rep[nx-1:0:-1, ny-1:0:-1, :nz+1    ]
        gint_rep[:nx+1, :ny+1, nz+1:] = gint_rep[:nx+1,     :ny+1,     nz-1:0:-1]
        gint_rep[nx+1:, :ny+1, nz+1:] = gint_rep[nx-1:0:-1, :ny+1,     nz-1:0:-1]
        gint_rep[:nx+1, ny+1:, nz+1:] = gint_rep[:nx+1,     ny-1:0:-1, nz-1:0:-1]
        gint_rep[nx+1:, ny+1:, nz+1:] = gint_rep[nx-1:0:-1, ny-1:0:-1, nz-1:0:-1]

        self._gint_rep = gint_rep.copy()

        # Tranasfer to device
        gint_rep_dev = context.nparray_to_context_array(gint_rep)

        # Prepare fft plan
        if fftplan is None:
            fftplan = context.plan_FFT(workspace_dev, axes=(0,1,2))

        # Transform the green function (in place)
        fftplan.transform(gint_rep_dev)

        self.dx = dx
        self.dy = dy
        self.dz = dz
        self.nx = nx
        self.ny = ny
        self.nz = nz
        self._workspace_dev = workspace_dev
        self._gint_rep_transf_dev = gint_rep_dev
        self.fftplan = fftplan

    #@profile
    def solve(self, rho):

        '''
        Solves Poisson's equation in free space for a given charge density.

        Args:
            rho (float64 array): charge density at the grid points in
                Coulomb/m^3.
        Returns:
            phi (float64 array): electric potential at the grid points in Volts.
        '''

        nz_alloc = self.nz
        if self._gint_rep_transf_dev.shape[2] > 1:
            nz_alloc = self._gint_rep_transf_dev.shape[2]
        _workspace_dev = self.context.zeros(
                (2*self.nx, 2*self.ny, nz_alloc), dtype=np.complex128, order='F')

        # The transposes make it faster in cupy (C-contigous arrays)
        _workspace_dev.T[:self.nz, :self.ny, :self.nx] = rho.T
        self.fftplan.transform(_workspace_dev) # rho_rep_hat

        try:
            _workspace_dev.T[:,:,:] *= (
                        self._gint_rep_transf_dev.T) # phi_rep_hat
        except Exception: # pyopencl does not support array broadcasting (used in 2.5D)
            for ii in range(self.nz):
                _workspace_dev.T[ii,:,:] *= (
                        self._gint_rep_transf_dev.T[0, :, :]) # phi_rep_hat

        self.fftplan.itransform(_workspace_dev) #phi_rep
        return _workspace_dev.real[:self.nx, :self.ny, :self.nz]

class FFTSolver2p5D(FFTSolver3D):

    '''
    Creates a Poisson solver object that solve's Poisson equation in
    the 2.5D aaoroximation equation the FFT method (free space).

    Args:
        nx (int): Number of cells in the horizontal direction.
        ny (int): Number of cells in the vertical direction.
        nz (int): Number of cells in the vertical direction.
        dx (float): Horizontal cell size in meters.
        dy (float): Vertical cell size in meters.
        dz (float): Longitudinal cell size in meters.
        context (XfContext): identifies the :doc:`context <contexts>`
            on which the computation is executed.
    Returns:
        (FFTSolver3D): Poisson solver object.
    '''

    def __init__(self, dx, dy, dz, nx, ny, nz, context=None, fftplan=None):

        if context is None:
            context = context_default
        self.context = context

        # Build grid for primitive function
        xg_F = np.arange(0, nx+2) * dx - dx/2
        yg_F = np.arange(0, ny+2) * dy - dy/2
        XX_F, YY_F= np.meshgrid(xg_F, yg_F, indexing='ij')

        # Compute primitive
        F_temp = primitive_func_2p5d(XX_F, YY_F)

        # Integrated Green Function (I will transform inplace)
        gint_rep= np.zeros((2*nx, 2*ny), dtype=np.complex128, order='F')
        gint_rep[:nx+1, :ny+1] = (F_temp[ 1:,  1:]
                                - F_temp[:-1,  1:]
                                - F_temp[ 1:, :-1]
                                + F_temp[:-1, :-1])

        # Replicate
        # To define how to make the replicas I have a look at:
        # np.abs(np.fft.fftfreq(10))*10
        # = [0., 1., 2., 3., 4., 5., 4., 3., 2., 1.]
        gint_rep[nx+1:, :ny+1] = gint_rep[nx-1:0:-1, :ny+1]
        gint_rep[:nx+1, ny+1:] = gint_rep[:nx+1, ny-1:0:-1]
        gint_rep[nx+1:, ny+1:] = gint_rep[nx-1:0:-1, ny-1:0:-1]


        # Prepare fft plan
        if fftplan is None:
            temp_dev = context.zeros((2*nx, 2*ny, nz),
                                    dtype=np.complex128, order='F')
            fftplan = context.plan_FFT(temp_dev, axes=(0,1))
            del(temp_dev)

        # Transform the green function
        gint_rep_transf = np.fft.fftn(gint_rep, axes=(0,1))

        # Transfer to GPU (if needed)
        gint_rep_transf_dev = context.nparray_to_context_array(
                                       np.atleast_3d(gint_rep_transf))

        self.dx = dx
        self.dy = dy
        self.dz = dz
        self.nx = nx
        self.ny = ny
        self.nz = nz
        self._gint_rep_transf_dev = gint_rep_transf_dev
        self.fftplan = fftplan


def primitive_func_3d(x,y,z):
    abs_r = np.sqrt(x * x + y * y + z * z)
    inv_abs_r = 1./abs_r
    res = 1./(4*pi*epsilon_0)*(
            -0.5 * (z*z * np.arctan(x*y*inv_abs_r/z)
                    + y*y * np.arctan(x*z*inv_abs_r/y)
                    + x*x * np.arctan(y*z*inv_abs_r/x))
               + y*z*np.log(x+abs_r)
               + x*z*np.log(y+abs_r)
               + x*y*np.log(z+abs_r))
    return res

def primitive_func_2p5d(x,y):

    abs_r = np.sqrt(x * x + y * y)
    inv_abs_r = 1./abs_r
    res = 1./(4*pi*epsilon_0)*(3*x*y - x*x*np.arctan(y/x)
                     - y*y*np.arctan(x/y) - x*y*np.log(x*x + y*y))
    return res
